@@ -18,6 +18,7 @@ open Structures
 [@@@ocaml.warning "-40"]
 
 module RO = Reductionops
+module C = Conversion
 module R = Reduction
 module EU = Evarutil
 module ES = Evarsolve
@@ -44,7 +45,8 @@ let set_debug b =
 let get_debug () = !debug
 
 let _ = Goptions.declare_bool_option {
-  Goptions.optdepr  = false;
+  Goptions.optdepr  = None;
+  Goptions.optstage = Interp;
   Goptions.optkey   = ["Unicoq";"Debug"];
   Goptions.optread  = get_debug;
   Goptions.optwrite = set_debug
@@ -52,7 +54,8 @@ let _ = Goptions.declare_bool_option {
 
 let trace = ref false
 let _ = Goptions.declare_bool_option {
-  Goptions.optdepr  = false;
+  Goptions.optdepr  = None;
+  Goptions.optstage = Interp;
   Goptions.optkey   = ["Unicoq";"Trace"];
   Goptions.optread  = (fun () -> !trace);
   Goptions.optwrite = (fun b -> trace := b);
@@ -60,7 +63,8 @@ let _ = Goptions.declare_bool_option {
 
 let dump = ref false
 let _ = Goptions.declare_bool_option {
-  Goptions.optdepr  = false;
+  Goptions.optdepr  = None;
+  Goptions.optstage = Interp;
   Goptions.optkey   = ["Unicoq";"Dump";"Equalities"];
   Goptions.optread  = (fun () -> !dump);
   Goptions.optwrite = (fun b -> dump := b);
@@ -68,7 +72,8 @@ let _ = Goptions.declare_bool_option {
 
 let latex_file = ref ""
 let _ = Goptions.declare_string_option {
-  Goptions.optdepr = false;
+  Goptions.optdepr = None;
+  Goptions.optstage = Interp;
   Goptions.optkey = ["Unicoq";"LaTex";"File"];
   Goptions.optread = (fun () -> !latex_file);
   Goptions.optwrite = (fun s-> latex_file := s);
@@ -111,14 +116,16 @@ let set_super_aggressive b =
       {!default_options with inst_super_aggressive = b}
 
 let _ = Goptions.declare_bool_option {
-  Goptions.optdepr = false;
+  Goptions.optdepr = None;
+  Goptions.optstage = Interp;
   Goptions.optkey = ["Unicoq"; "Aggressive"];
   Goptions.optread = is_aggressive;
   Goptions.optwrite = set_aggressive;
 }
 
 let _ = Goptions.declare_bool_option {
-  Goptions.optdepr = false;
+  Goptions.optdepr = None;
+  Goptions.optstage = Interp;
   Goptions.optkey = ["Unicoq"; "Super"; "Aggressive"];
   Goptions.optread = is_super_aggressive;
   Goptions.optwrite = set_super_aggressive;
@@ -131,7 +138,8 @@ let set_solving_eqn b =
     {!default_options with inst_try_solving_eqn = b}
 
 let _ = Goptions.declare_bool_option {
-  Goptions.optdepr  = false;
+  Goptions.optdepr  = None;
+  Goptions.optstage = Interp;
   Goptions.optkey   = ["Unicoq"; "Try"; "Solving";"Eqn"];
   Goptions.optread  = get_solving_eqn;
   Goptions.optwrite = set_solving_eqn
@@ -144,7 +152,8 @@ let set_hash b =
     {!default_options with use_hash = b}
 
 let _ = Goptions.declare_bool_option {
-  Goptions.optdepr = false;
+  Goptions.optdepr = None;
+  Goptions.optstage = Interp;
   Goptions.optkey = ["Unicoq"; "Use";"Hash"];
   Goptions.optread = use_hash;
   Goptions.optwrite = set_hash;
@@ -299,7 +308,7 @@ let log_eq env rule conv_t t1 t2 (l, sigma) =
     let ppcmd_of env (t : EConstr.t) =
       try Printer.pr_econstr_env env sigma t
       (* This is really suspicious as it will hide a serious bug. *)
-      with _ -> Termops.Internal.debug_print_constr t
+      with _ -> Constr.debug_print (EConstr.Unsafe.to_constr t)
     in
     let str1 = Pp.string_of_ppcmds (ppcmd_of env t1) in
     let str2 = Pp.string_of_ppcmds (ppcmd_of env t2) in
@@ -326,14 +335,14 @@ let debug_str s _l =
 let debug_eq env sigma t c1 c2 _l =
   let s1 = string_of_ppcmds (Printer.pr_econstr_env env sigma (applist c1)) in
   let s2 = string_of_ppcmds (Printer.pr_econstr_env env sigma (applist c2)) in
-  Printf.printf "%s %s %s\n" s1 (if t == R.CONV then "=?=" else "<?=") s2;
+  Printf.printf "%s %s %s\n" s1 (if t == C.CONV then "=?=" else "<?=") s2;
   flush_all ()
 
 let print_eq f (conv_t, c1, c2) =
   output_string f "\\lstinline{";
   output_string f c1;
   output_string f "}~\\approx_{";
-  output_string f (if conv_t == R.CONV then "=" else "\\leq");
+  output_string f (if conv_t == C.CONV then "=" else "\\leq");
   output_string f "}~\\lstinline{";
   output_string f c2;
   output_string f "}"
@@ -442,7 +451,7 @@ let get_definition sigma env t : EConstr.t =
     unfolds c and returns the new head term d applied to the concatenation
     of arguments. *)
 let get_def_app_stack sigma env (c, args) =
-  let (d, dargs) = decompose_app sigma (get_definition sigma env c) in
+  let (d, dargs) = decompose_app_list sigma (get_definition sigma env c) in
   (d, dargs @ args)
 
 let try_unfolding sigma ts env t =
@@ -713,7 +722,7 @@ let evar_apprec ts env sigma (c, stack) =
       aux (Evd.existential_value sigma (evk,eva), stack)
     | _ ->
       match RO.Stack.list_of_app_stack stack with
-      | None -> decompose_app sigma (RO.Stack.zip sigma (t, stack))
+      | None -> decompose_app_list sigma (RO.Stack.zip sigma (t, stack))
       | Some stack -> (t, stack)
   in aux (c, RO.Stack.append_app_list stack RO.Stack.empty)
 
@@ -771,18 +780,18 @@ end
 
 type unify_fun =
   Environ.env -> Evd.evar_map ->
-  Reduction.conv_pb -> EConstr.t -> EConstr.t -> Evarsolve.unification_result
+  Conversion.conv_pb -> EConstr.t -> EConstr.t -> Evarsolve.unification_result
 
 (** The main module type of unification, containing the functions that can be exported *)
 module type Unifier = sig
   val unify_evar_conv : unify_fun
 
-  val unify_constr : ?conv_t:R.conv_pb ->
+  val unify_constr : ?conv_t:C.conv_pb ->
                      ?options:options ref ->
                      Environ.env ->
                      EConstr.t -> EConstr.t -> Logger.log * Evd.evar_map -> unif
 
-  val instantiate : R.conv_pb ->
+  val instantiate : C.conv_pb ->
            options ref ->
            Environ.env ->
            EConstr.t Constr.pexistential * EConstr.t list ->
@@ -834,7 +843,7 @@ module Inst = functor (U : Unifier) -> struct
       let sigma = prune_all map sigma0 in
       let sigma, t' =
 	Evarsolve.refresh_universes
-	    (if conv_t == R.CUMUL && isArity sigma t' then
+            (if conv_t == C.CUMUL && isArity sigma t' then
 	      (* ?X <= Type(i) -> X := Type j, j <= i *)
 	      (* Type(i) <= X -> X := Type j, i <= j *)
 	      Some (dir == Original)
@@ -851,9 +860,11 @@ module Inst = functor (U : Unifier) -> struct
 	       The body of ?X and ?Y just has to be of type Π Δ. Type k for some k <= i, j. *)
 	    let evienv = Evd.evar_env env evi in
 	    let ctx1, i = R.dest_arity evienv (EConstr.to_constr ~abort_on_undefined_evars:false sigma (Evd.evar_concl evi)) in
-	    let evi2 = Evd.find sigma evk2 in
+	    let evi2 = Evd.find_undefined sigma evk2 in
 	    let evi2env = Evd.evar_env env evi2 in
 	    let ctx2, j = R.dest_arity evi2env (EConstr.to_constr ~abort_on_undefined_evars:false sigma (Evd.evar_concl evi2)) in
+	    let i = ESorts.make i in
+	    let j = ESorts.make j in
 	    if i == j || Evd.check_eq sigma i j then (* Shortcut, i = j *)
 	      (dbg, ES.Success sigma)
 	    else if Evd.check_leq sigma i j then
@@ -872,7 +883,7 @@ module Inst = functor (U : Unifier) -> struct
 	  with R.NotArity ->
             if !options.inst_unify_types then begin
                 let ty' = Retyping.get_type_of env sigma t'' in
-                U.unify_constr ~conv_t:R.CUMUL env ty' ty (dbg, sigma)
+                U.unify_constr ~conv_t:C.CUMUL env ty' ty (dbg, sigma)
               end
             else
               (dbg, ES.Success sigma)
@@ -924,7 +935,7 @@ module struct
   let try_hash env sp1 sp2 (dbg, sigma as dsigma) =
     if use_hash () && tblfind tbl (sigma, env, sp1, sp2) then
       begin
-        log_eq_spine env "Hash-Hit" R.CONV sp1 sp2 dsigma &&= fun (dbg, sigma) ->
+        log_eq_spine env "Hash-Hit" C.CONV sp1 sp2 dsigma &&= fun (dbg, sigma) ->
           report (dbg, ES.UnifFailure (sigma, PE.NotSameHead))
       (* TODO Something is funky here: it returns error on a success
          and the other way around?? *)
@@ -960,7 +971,7 @@ module struct
       an evar, returning the new head and list of arguments.
   *)
   let rec decompose_evar sigma (c, l) =
-    let (c', l') = decompose_app sigma c in
+    let (c', l') = decompose_app_list sigma c in
     if isCast sigma c' then
       let (t, _, _) = destCast sigma c' in
       decompose_evar sigma (t, l' @ l)
@@ -968,7 +979,7 @@ module struct
       (c', l' @ l)
 
   (** {3 "The Function" is split into several} *)
-  let rec unify'' ?(conv_t=R.CONV) ?(options=default_options) cont env t t' (dbg, sigma) =
+  let rec unify'' ?(conv_t=C.CONV) ?(options=default_options) cont env t t' (dbg, sigma) =
     assert (not (isApp sigma (fst t) || isApp sigma (fst t')));
     let (c, l as t) = decompose_evar sigma t in
     let (c', l' as t') = decompose_evar sigma t' in
@@ -990,11 +1001,11 @@ module struct
           Hashtbl.add tbl (sigma, env, t, t') true;
         res
 
-  and unify' ?(conv_t=R.CONV) ?(options=default_options) =
+  and unify' ?(conv_t=C.CONV) ?(options=default_options) =
     unify'' ~conv_t ~options (fun conv_t env t t' sigma dbg -> try_step conv_t env t t' (dbg, sigma))
 
-  and unify_constr ?(conv_t=R.CONV) ?(options=default_options) env t t' (dbg, sigma) =
-    unify' ~conv_t ~options env (decompose_app sigma t) (decompose_app sigma t') (dbg,sigma)
+  and unify_constr ?(conv_t=C.CONV) ?(options=default_options) env t t' (dbg, sigma) =
+    unify' ~conv_t ~options env (decompose_app_list sigma t) (decompose_app_list sigma t') (dbg,sigma)
 
   and unify_evar_conv env sigma0 conv_t t t' =
     let interesting log = (* if it's not just Reduce-Same *)
@@ -1035,9 +1046,9 @@ module struct
 
   and run_and_unify dbg env sigma0 args ty =
     let a, f, v = List.nth args 0, List.nth args 1, List.nth args 2 in
-    unify' ~conv_t:R.CUMUL env (decompose_app sigma0 a) ty (dbg, sigma0) &&= fun (dbg, sigma1) ->
+    unify' ~conv_t:C.CUMUL env (decompose_app_list sigma0 a) ty (dbg, sigma0) &&= fun (dbg, sigma1) ->
       match !run_function env sigma1 f with
-      | Some (sigma2, v') -> unify' env (decompose_app sigma2 v) (decompose_app sigma2 v') (dbg, sigma2)
+      | Some (sigma2, v') -> unify' env (decompose_app_list sigma2 v) (decompose_app_list sigma2 v') (dbg, sigma2)
       | _ -> (dbg, ES.UnifFailure (sigma1, PE.NotSameHead))
 
   and try_canonical_structures env (c, _ as t) (c', _ as t') sigma dbg =
@@ -1065,10 +1076,10 @@ module struct
         (evd,[],List.length bs) bs
     in
     report (
-      log_eq_spine env "CS" R.CONV t t' (dbg, evd') &&=
+      log_eq_spine env "CS" C.CONV t t' (dbg, evd') &&=
       ise_list2 (fun x1 x -> unify_constr env x1 (substl ks x)) params1 params &&=
       ise_list2 (fun u1 u -> unify_constr env u1 (substl ks u)) us2 us &&=
-      unify' env (decompose_app evd' c1) (c,(List.rev ks)) &&=
+      unify' env (decompose_app_list evd' c1) (c,(List.rev ks)) &&=
       ise_list2 (unify_constr env) ts ts1)
 
   and try_app_fo conv_t env (c, l as t) (c', l' as t') sigma dbg =
@@ -1147,13 +1158,13 @@ module struct
         (* TODO: Why the [] here!! *)
         let t = ES.solve_pattern_eqn env sigma [] (applist t) in
         let pbty = match conv_t with
-	    R.CONV -> None
-          | R.CUMUL -> Some (dir == Original)
+            C.CONV -> None
+          | C.CUMUL -> Some (dir == Original)
         in
         match ES.solve_simple_eqn (fun _flags _k -> unify_evar_conv) P.flags env sigma (pbty, evsubs, t) with
           ES.Success sigma' ->
           Printf.printf "%s" "solve_simple_eqn solved it: ";
-	  debug_eq env sigma R.CONV (mkEvar evsubs, []) (decompose_app sigma t) 0;
+          debug_eq env sigma C.CONV (mkEvar evsubs, []) (decompose_app_list sigma t) 0;
 	  (dbg, ES.Success sigma')
 	| e -> (dbg, e)
       with _ ->
@@ -1171,8 +1182,8 @@ module struct
         begin
           try
 	    let sigma1 = match conv_t with
-              | R.CONV -> Evd.set_eq_sort env sigma0 (ESorts.kind sigma0 s1) (ESorts.kind sigma0 s2)
-	      | R.CUMUL -> Evd.set_leq_sort env sigma0 (ESorts.kind sigma0 s1) (ESorts.kind sigma0 s2)
+              | C.CONV -> Evd.set_eq_sort env sigma0 s1 s2
+              | C.CUMUL -> Evd.set_leq_sort env sigma0 s1 s2
             in
             report (dbg, ES.Success sigma1)
           with UGraph.UniverseInconsistency e ->
@@ -1260,14 +1271,14 @@ module struct
     match (kind sigma0 c, kind sigma0 c') with
     (* Lam-BetaR *)
     | _, Lambda (_, _, trm) when reduce_right && not (CList.is_empty l') ->
-      let (c2, l2) = decompose_app sigma0 (subst1 (List.hd l') trm) in
+      let (c2, l2) = decompose_app_list sigma0 (subst1 (List.hd l') trm) in
       let t2 = (c2, l2 @ List.tl l') in
       report (
         log_eq_spine env "Lam-BetaR" conv_t t t' (dbg, sigma0) &&=
         unify' ~conv_t env t t2)
 
     | _, LetIn (_, trm, _, body) when reduce_right ->
-      let (c2, l2) = decompose_app sigma0 (subst1 trm body) in
+      let (c2, l2) = decompose_app_list sigma0 (subst1 trm body) in
       let t2 = (c2, l2 @ l') in
       report (
         log_eq_spine env "Let-ZetaR" conv_t t t' (dbg, sigma0) &&=
@@ -1286,14 +1297,14 @@ module struct
 
     (* Lam-BetaL *)
     | Lambda (_, _, trm), _ when reduce_left && not (CList.is_empty l) ->
-      let (c1, l1) = decompose_app sigma0 (subst1 (List.hd l) trm) in
+      let (c1, l1) = decompose_app_list sigma0 (subst1 (List.hd l) trm) in
       let t1 = (c1, l1 @ List.tl l) in
       report (
         log_eq_spine env "Lam-BetaL" conv_t t t' (dbg, sigma0) &&=
         unify' ~conv_t env t1 t')
     (* Let-ZetaL *)
     | LetIn (_, trm, _, body), _ when reduce_left ->
-      let (c1, l1) = decompose_app sigma0 (subst1 trm body) in
+      let (c1, l1) = decompose_app_list sigma0 (subst1 trm body) in
       let t1 = (c1, l1 @ l) in
       report (
         log_eq_spine env "Let-ZetaL" conv_t t t' (dbg, sigma0) &&=
@@ -1552,7 +1563,7 @@ module struct
     let sigma', univ = Evd.new_sort_variable Evd.univ_flexible sigma in
     let sigma'',v = Evarutil.new_pure_evar (EConstr.val_of_named_context nc') sigma' (EConstr.mkSort univ) in
     let idsubst = (mkRel 1 :: id_substitution nc) in
-    unify_constr ~conv_t:R.CUMUL env ty
+    unify_constr ~conv_t:C.CUMUL env ty
       (mkProd (Context.make_annot (Names.Name naid) Sorts.Relevant, a, mkLEvar sigma (v, idsubst)))
       (dbg, sigma'')
 
@@ -1620,7 +1631,7 @@ let unify_match_nored evars ts =
   let module M = (val unif (module P)) in
   M.unify_evar_conv
 
-let instantiate ?(conv_t=R.CONV) ?(options=default_options) env
+let instantiate ?(conv_t=C.CONV) ?(options=default_options) env
       (_, _ as evsubs) t sigma =
   let module P = (struct
     let flags = Evarconv.default_flags_of TransparentState.full
@@ -1629,7 +1640,7 @@ let instantiate ?(conv_t=R.CONV) ?(options=default_options) env
     let match_evars = None
   end : Params) in
   let module M = (val unif (module P)) in
-  M.instantiate conv_t options env evsubs (decompose_app sigma t) sigma
+  M.instantiate conv_t options env evsubs (decompose_app_list sigma t) sigma
 
 let use_munify () = !munify_on
 let set_use_munify b =
@@ -1637,7 +1648,8 @@ let set_use_munify b =
   munify_on := b
 
 let _ = Goptions.declare_bool_option {
-  Goptions.optdepr = false;
+  Goptions.optdepr = None;
+  Goptions.optstage = Interp;
   Goptions.optkey = ["Use";"Unicoq"];
   Goptions.optread = use_munify;
   Goptions.optwrite = set_use_munify;
